#!/usr/bin/env python3
# vim: set ft=python fileencoding=utf-8 tw=72:
r"""
A blog-oriented static web content generator.

* Input formats: Markdown and reStructuredText
* Output format: HTML5.

:author: \J. A. Corbal
:email: jacorbal@gmail.com
:copyright: © 2012-2020, J. A. Corbal
:license: MIT
"""

from pynfact.main import main


if __name__ == '__main__':
    main()



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Markdown translation to HTML and metadata retrieval.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT
"""
from pynfact.fileman import has_extension_md, has_extension_rst
from pynfact.parsers import ParserMd
from pynfact.parsers import ParserRst


class Parser:
    """Parse methods for a markup language file to generate HTML code.

    .. versionchanged:: 1.3.1b1
        Former class ``Mulang``, now it relies on the file extension to
        call one or other parser.
    """

    def __init__(self, input_data, encoding='utf-8', logger=None):
        """Constructor.

        Depending on the extension, the markup parser will be a Markdown
        parser, or a reStructuredText parser.  The valid extensions are:
        ``.rst`` for reStructuredText, and ``.md`` for Markdown.

        :param input_data: File from where the data is taken
        :type input_data: str
        :param encoding: Encoding the input file is in
        :type encoding: str
        :param logger: Logger where to store activity in
        :type logger: logging.Logger
        """
        # Parser = ParserRst if has_extension_rst(input_data) else ParserMd
        if has_extension_rst(input_data):
            Parser = ParserRst
        elif has_extension_md(input_data):
            Parser = ParserMd

        self.parser = Parser(input_data, encoding, logger=logger)

    def html(self):
        """Generate HTML from a MarkUP LANGuage file."""
        return self.parser.html()

    def metadata(self):
        """Generate metadata from a MarkUP LANGuage file."""
        return self.parser.metadata()



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Data and data structures management functions.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT

.. versionadded:: 1.3.1a3
    Separate funcionality across multiple modules.

.. versionchanged:: 1.3.1b1
    Remove funcion ``and_array_in``: it has no use for this project.
"""


def or_array_in(dictionary, *keys):
    """Check if at least one key in the array is in the dictionary.

    The funcionality is the same as the keyword ``in``, but testing each
    and every key in the array as an ``or`` operation.

    :param keys: List of keys to test against the dictionary
    :type keys: args
    :param dictionary: Dictionary
    :type dictionary: dict
    :return: Value for the matching key, or ``None``
    :rtype: mixed

    :Example:

    >>> dictionary = {'a': 1, 'b': 2, 'c': 3, 'd': 4}

    >>> or_array_in(dictionary, ['a', 'i', 'j']
    True

    >>> or_array_in(dictionary, ['x', 'y', 'z']
    False
    """
    for key in keys:
        if key in dictionary:
            return dictionary.get(key)
    return None


def key_exists(dictionary, *keys):
    """Check if key exist in a nested dictionary.

    :param dictionary: Nested dicionary
    :type dictionary: dict
    :param keys: Keys to test agains the dictionary
    :type keys: args
    :return: ``True`` if at least one key is in the dicionary
    :rtype: bool
    :raise KeyError: If no key is in the dictionary
    """
    item = dictionary
    for key in keys:
        try:
            item = item[key]
        except KeyError:
            return False
    return True



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
File and path manipulation functions.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT
"""

from pynfact.struri import slugify, strip_html_tags
import os


def has_extension(filename, extension, case_sensitive=False):
    """Test if the filename has a extension in particular.

    :param filename: Filename to test
    :type filename: str
    :param extension: Extension to test against
    :type extension: str
    :return: ``True`` if the filename has the inquired extension
    :rtype: bool

    :Example:

    >>> has_extension('filename.MD', '.md', case_sensitive=True)
    False

    >>> has_extension('filename.RST', '.rst', case_sensitive=False)
    True

    >>> has_extension('filename.txt', '.md')
    False
    """
    if case_sensitive:
        return os.path.splitext(filename)[1] == extension
    else:
        return os.path.splitext(filename)[1].lower() == \
            extension.lower()


def has_extensions(filename, extensions, case_sensitive=False):
    """Check if a filename has an extension in a list of extensions.

    :param filename: Filename to test
    :type filename: str
    :param extensions: List of extensions to test against
    :type extensions: list
    :param case_sensitive: Do not considerate case if ``False``
    :type case_sensitive: bool
    :return: ``True`` if filename has a matching extension
    :rtype: bool

    :Example:

    >>> has_extensions('filename.md', ['.md', '.rst'])
    True

    >>> has_extensions('filename.rst', ['.md', '.rst'])
    True

    >>> has_extensions('filename.txt', ['.md', '.rst'])
    False
    """
    if case_sensitive:
        return os.path.splitext(filename)[1] in extensions
    else:
        return os.path.splitext(filename)[1].lower() in \
            [extension.lower() for extension in extensions]


def has_extension_md(filename):
    """Check whether a filename has a valid Markdown extension.

    :Example:

    >>> has_extension_md('filename.md')
    True

    >>> has_extension_md('filename.mDown')
    True

    >>> has_extension_md('filename.rst')
    False

    """
    return has_extensions(filename, ['.md', '.mdown', '.mkdn', '.markdown'])


def has_extension_rst(filename):
    """Check whether a filename has a valid reStructuredText extension.

    :Example:

    >>> has_extension_rst('filename.md')
    False

    >>> has_extension_rst('filename.rst')
    True

    >>> has_extension_rst('filename.reSt')
    True

    """
    return has_extensions(filename, ['.rst', '.rest', '.rtext'])


def has_extension_md_rst(filename):
    """Check if a filename has Markdown or reStructuredText extension.

    :Example:

    >>> has_extension_md_rst('filename.md')
    True

    >>> has_extension_md_rst('filename.rst')
    True

    >>> has_extension_rst('filename.txt')
    False
    """
    return has_extension_md(filename) or has_extension_rst(filename)


def link_to(name, prefix='', makedirs=True, justdir=False,
            index='index.html'):
    """Make a link relative path in terms of the build.

    It may return a string to the destination, and also create that path
    if it doesn't exist.

    :param name: Filename (extension will be disregarded)
    :type name: str
    :param prefix: Prefix directory to prepend the file
    :type prefix: str
    :param makedirs: If true make all needed directories
    :type makedirs: bool
    :param justdir: If true return only the link dir., not the full path
    :type justdir: bool
    :param index: Default name of file contained in the path
    :type index: str
    :return: Link path to a  ``index.html`` page
    :rtype: str

    :Example:

    >>> struri.link_to('dest', '/a/b/c', makedirs=False)
    '/a/b/c/dest/index.html'

    >>> struri.link_to('dest', 'a/b/c', makedirs=False, justdir=True)
    'a/b/c/dest'

    .. versionchanged:: 1.2.0a1
        Relocate function location from module :mod:`struri` to
        :mod:`fileman`.

    .. todo::
        Add a verbose argument, default to ``False`` (for
        compatibility), that prints status on creating the path if
        ``makedirs=True``.
    """
    dirname = slugify(strip_html_tags(os.path.splitext(name)[0]))
    path = os.path.join(prefix, slugify(dirname), index)
    if makedirs:
        os.makedirs(os.path.dirname(path), exist_ok=True)

    return os.path.dirname(path) if justdir else path



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Markdown translation to HTML and metadata retrieval.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT
"""
import markdown


class ParserMd:
    """Generate HTML body from a Markdown source file.

    .. versionchanged:: 1.2.0a1
        Implement ``logging`` instead of printing to ``stdout`` and/or
        ``stderr``.

    .. versionchanged:: 1.3.1a4
        This class, formerly ``Mulang``, now it's just a parser for
        Markdown since reStrucutedText support was added.

    .. seealso:: :class:`ParserRst` and :mod:`parser`.
    """

    def __init__(self, input_data, encoding='utf-8', logger=None):
        """Constructor.

        This class has methods to parse the markup language and get the
        information.  The logging activity is only set to debug, since
        it's not necessary to tell continuously the default activity.

        :param input_data: File from where the data is taken
        :type input_data: str
        :param encoding: Encoding the input file is in
        :type encoding: str
        :param logger: Logger where to store activity in
        :type logger: logging.Logger
        """
        self.input_data = input_data
        self.encoding = encoding
        self.logger = logger

        self.md = markdown.Markdown(
            extensions=['markdown.extensions.extra',
                        'markdown.extensions.toc',
                        'markdown.extensions.abbr',
                        'markdown.extensions.def_list',
                        'markdown.extensions.footnotes',
                        'markdown.extensions.codehilite',
                        'markdown.extensions.meta'],
            encoding=encoding,
            output_format='html5')

    def html(self):
        """Generate HTML from a Markdown file."""
        with open(self.input_data, "r", encoding=self.encoding) as f:
            text = f.read()

        html = self.md.convert(text)
        self.logger and self.logger.debug(
            'Parsed body data of: "{}"'.format(self.input_data))

        return html

    def metadata(self):
        """Fetch metadata in a Markdown file.

        The metadata syntax is case insensitive and as follows::

            author: Author A. Author
            title: This is a title
            subtitle: A longer, much longer title for this entry
            cdate: 2020-03-10
            mdate: 2020-03-19
            tags: this, are, a, lot, of, tags
            comments: yes

        .. todo::
            Get the meta without generating HTML again.
        """
        with open(self.input_data, "r", encoding=self.encoding) as f:
            text = f.read()

        html = self.md.convert(text)
        self.logger and self.logger.debug(
            'Parsed meta data of: "{}"'.format(self.input_data))

        return self.md.Meta



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
reStructuredText translation to HTML and metadata retrieval.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT

.. note::
    There are some presentation and parsing probles using the docutils
    ``Writer`` class from ``html5_polyglot`` if docutils version is
    ``<0.15``.  The result is not the same, some settings are ignored,
    and the ``doctree`` seems to be interpreted in a different way
    concerning the ``docinfo`` element, way too important for the
    document metadata.  The ``fragment`` and the ``body`` parts of the
    publisher data structure seem to have also some erratic behaviour.
    It's required ``docutils>=0.15``.
"""

from docutils.core import publish_string, publish_parts, publish_doctree
from docutils.nodes import docinfo
from docutils.writers.html5_polyglot import Writer


class ParserRst:
    """Generate HTML body from a reStructuredText source file.

    .. versionadded:: 1.3.1a4
        Add reStructuredText support.

    .. seealso:: :class:`ParserMd` and :mod:`parser`.
    """

    def __init__(self, input_data, encoding='utf-8', logger=None):
        """Constructor.

        This class has methods to parse the markup language and get the
        information.  The logging activity is only set to debug, since
        it's not necessary to tell continuously the default activity.

        :param input_data: File from where the data is taken
        :type input_data: str
        :param encoding: Encoding the input file is in
        :type encoding: str
        :param output_format: Output HTML version
        :type output: str
        :param logger: Logger where to store activity in
        :type logger: logging.Logger
        """
        self.input_data = input_data
        self.encoding = encoding
        self.logger = logger

        self.settings = {
            'attribution': None,  # dash, parentheses
            'cloak_email_addresses': True,
            'compact_field_lists': True,
            'compact_lists': True,
            'doctitle_xform': True,
            'embed_stylesheet': False,
            'footnote_references': 'superscript',  # 'brackets'
            'initial_header_level': 2,
            'input_encoding': self.encoding,
            'language_code': 'en',
            'math_output': 'MathML',  # 'LaTeX', 'MathJax', 'HTML'
            'output_encoding': 'utf-8',
            'raw_enabled': True,
            'sectsubtitle_xform': True,
            'smart_quotes': False,
            'smartquotes_locales': True,
            'strip_comments': True,
            # 'strip_elements_with_classes': ['docinfo simple', 'docinfo'],
            # 'stylesheet_path': stylesheet_path,
            'syntax_highlight': True,
            'table_style': 'align-center',  # borderless, booktabs, align-
            'tab_width': 4,
            'toc_backlinks': False,
            'trim_footnote_reference_space': True,
        }

    def html(self):
        """Generate HTML from  a reStructuredText file.

        .. todo::
            This reStructuredText parser should import some locally
            defined roles for this application.
        """
        with open(self.input_data, "r", encoding=self.encoding) as f:
            text = f.read()

        html = publish_parts(text,
                             writer=Writer(),
                             settings_overrides=self.settings).get('fragment')
        self.logger and self.logger.debug(
            'Parsed body data of: "{}"'.format(self.input_data))

        return html

    def metadata(self):
        """Fetch metadata in a reStructuredText file.

        The metadata syntax is case insensitive and as follows::

            :author: Author A. Author
            :title: This is a title
            :subtitle: A longer, much longer title for this entry
            :cdate: 2020-03-10
            :mdate: 2020-03-19
            :tags: this, are, a, lot, of, tags
            :comments: yes

        .. note::
            This is a "cheap" approximation that I dislike.  The
            non-bibliographical fields are gotten in a different way
            than the bibliographical ones.  In docutils, the
            bibliographical fields are::

                address, author, authors, contact, copyright, date,
                field, organization, revision, status, version

            and are processed by docutils in a different way that user
            defined fields.  In order to make a dictionary that does not
            treat any field differently, we check if a field is
            biliographic (length=1), or not, (length=2).  If not, we
            store normally, if it is, we take the short representation
            and strip it from its brackets and ellipsis.

        .. todo::
            Fix how to get the fields in a much more cleaner way.
        """
        meta = dict()
        with open(self.input_data, "r", encoding=self.encoding) as f:
            text = f.read()

        doctree = publish_doctree(text,  # writer=Writer(),
                                  settings_overrides=self.settings)

        # Generate dictionary of meta information {'field': 'value'}
        for info in doctree.traverse(docinfo):
            for field in info.children:
                if len(field.children) == 1:
                    k, v = field.shortrepr().strip('<.>'), field
                    meta[str(k).lower()] = v.astext()
                elif len(field.children) == 2:
                    k, v = field.children
                    meta[k.astext().lower()] = v.astext()

        # Lowercase all the keys in the docinfo, and use the same
        # structure `{k: [v]}` as the Markdown metadata retriever
        meta = {k.lower(): [v] for k, v in meta.items()}

        # In reStructuredText, if there's no title in the meta tags, get
        # it from the document part 'title'
        if 'title' not in meta:
            meta['title'] = parts.get('title')

        self.logger and self.logger.debug(
            'Parsed body data of: "{}"'.format(self.input_data))

        return meta



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
URI strings manipulation functions.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT

.. versionchanged:: 1.2.0a1
    Relocate file management functions to :mod:`fileman`.
"""
from datetime import datetime
from dateutil import tz
import unidecode
import re


def slugify(unslugged, separator='-'):
    """Slug a string.

    :param unslugged: String to slugify
    :type unslugged: str
    :param separator: Character used to separate words
    :type separator: str
    :return: Slugged string
    :rtype: str

    :Example:

    >>> slugify('Here you go a not so very long string')
    'here-you-go-a-not-so-very-long-string'

    >>> slugify('Does**this%&string=has--DASHES!?, [Also] symbols?')
    'does-this-string-has-dashes-also-symbols'

    .. versionchanged:: 1.3.1b3
        Strip beginning and ending separator characters.
    """
    return re.sub(r'\W+', separator,
                  unidecode.unidecode(
                      unslugged).strip().lower()).strip(separator)


def strip_html_tags(text):
    """Strip HTML tags in a string.

    :param text: String containing HTML code
    :type text: str
    :return: String without HTML tags
    :rtype: str

    :Example:

    >>> strip_html_tags('<div><p>This is a paragraph</div>')
    'This is a paragraph'

    >>> strip_html_tags('<em class="highlight">Highlighted</em> text')
    'Highlighted text'    
    """
    return re.sub('<[^<]+?>', '', text)


def date_iso(date):
    """Convert a datetime string into ISO 8601 format.

    HTML date format agrees with ISO 8601 (see also, :RFC:`3339`), ie::

        YYYY[-MM[-DD]][Thh[:mm[:ss[.s]]]T]

    For more information:

    * `Date and Time Formats:
      <https://www.w3.org/TR/NOTE-datetime>`_
    * `Date formats:
      <https://www.w3.org/International/questions/qa-date-format>`_

    :param date: Datetime object
    :type date: datetime.datetime
    :return: Datetime formatted as ISO 8601, or empty string if invalid
    :rtype: str

    .. note::
        If the datetime object it timezone naïve, it'll be localized to
        UTC, so the feed parser, and any other function that requires
        timezone aware datetime objects, do not raise an exception.
    """
    if date and not datetime.strftime(date, '%z'):
        date = date.replace(tzinfo=tz.tzutc())
    return datetime.isoformat(date, timespec='minutes') if date else ''



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Meta information processor from Markdown or reStructuredText file.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT

.. note::
    In the metainformation header of every Markdown document to be
    parsed, there may be diacritics in metainformation values (UTF-8 by
    default) but not in metainformation key names (``[A-Za-z0-9]]``).

.. versionchanged:: 1.3.1a3
    Meta tags in document header are now reduced to only English,
    although the code is set to add more.  It's possible that in next
    versions those fields will be added by importing a JSON file will
    all the translations, but for the time being, it's better to stick
    to one universal language instead of hardcoding the
    internationalization.
"""
from dateutil.parser import parse as dt_parse
from pynfact.dataman import or_array_in
from pynfact.fileman import has_extensions
from pynfact.struri import date_iso, strip_html_tags
import logging
import markdown
import re
import sys
import locale


class Meta:
    """Meta information processor.

    The ``Meta`` class also validates the metadata with the method
    :func:``Meta._is_valid``.  It seems off at first glance for a
    metadata retriever class, but it's responsability of this class
    ensure that the metadata is well-formed.

    .. versionchanged:: 1.3.1a3
        Validate source input meta tags of title (always) and date (when
        needed) raising a ``ValueError`` and passing to the logger the
        information of the error.

    .. versionchanged:: 1.3.1b1
        Filetype no longer required in the constructor, since the parser
        automatically detect the file type by checking its extension.

    .. note::
        The prefixes for date are ``o-`` for "original date", or the
        date when it was created (and not ``c-``, commonly used for
        "changed date"); and ``m-`` for modified date".
    """

    def __init__(self, meta, filename='', odate_required=False, logger=None):
        """Constructor.

        :param meta: Meta information dictionary
        :type meta: dict
        :param filename: Identifier for the file that is being tested
        :type filename: str
        :param odate_required: If ``True`` checks for the date metadata
        :type odate_required: bool
        :param logger: Logger where to store activity in
        :type logger: logging.Logger
        """
        self.meta = meta
        self.filename = filename
        self.odate_required = odate_required
        self.logger = logger
        self._is_valid()  # Validate data supplied

    def as_dict(self, override_values={}, defaults={}, date_format='%c'):
        """Return an array with all metadata.

        :param override_values: Dictionary of values to update
        :type:override_values: dict
        :param defaults: Default values if metadata value is empty
        :type defaults: dict
        :param date_format: Datetime format strings
        :type date_format: str
        :return: Dictionary of all metadata
        :rtype: dict
        """
        metadata = {
            'author': defaults['author']
                if not self.author()
                else self.author(),
            'category': defaults['category']
                if not self.category()
                else self.category(),
            'comments': self.comments(),
            'copyright': self.copyright(),
            'email': self.email(),
            'language': defaults['language']
                if not self.language()
                else self.language(),
            'mdate_html': date_iso(self.mdate_info()),
            'mdate_info': self.mdate_info(),
            'mdate': self.mdate(date_format),
            'odate_html': date_iso(self.odate_info()),
            'odate_info': self.odate_info(),
            'odate': self.odate(date_format),
            'navigation': self.navigation(),
            'private': self.private(),
            'raw_title': strip_html_tags(self.title()),
            'subtitle': self.subtitle(),
            'tag_list': self.tag_list(),
            'title': self.title(),
        }

        return metadata.update(override_values) or metadata

    def title(self):
        """Get post title as a string from post meta information.

        This function performs some validation procedures in order to
        make sure that the title is valid (ie, not empty).

        .. note::
            In order to make the title a mandatory field, it's required
            to have the ``default`` argument to ``None``, so it raises
            an exception to warn the user.

        :param default: Default value if title is empty
        :type default: str
        :return: Title field
        :rtype: str
        :raise ValueError: If the title field is empty
        """
        vs = {'title'}
        rtitle = self._parse_str_md(vs, None, ' ')

        try:
            if not rtitle:
                raise ValueError
        except ValueError:
            self.logger and self.logger.error(
                'Empty title value in "{}" metadata'.format(self.filename))
            sys.exit(33)
        else:
            return rtitle

    def subtitle(self, default=''):
        """Get post subtitle as a string from post meta information.

        Add a full stop or period at the end if the parsed string
        doesn't have one.

        :param default: Default value if title is empty
        :type default: str
        :return: Summary field
        :rtype: str
        """
        vs = {'subtitle', 'summary'}
        return self._parse_str_md(vs, default, ' ', period=True)

    def category(self, default='Miscellaneous'):
        """Get post category as a string from post meta information.

        .. note::
            A post can belong to only one category.  If the user inputs
            more than one category, the first one will be the valid one.

        :param default: Default value if category wasn't specified
        :type default: str
        :return: Value of the category field
        :rtype: str
        """
        vs = {'category', 'cat'}
        return self._parse_str_raw(vs, default, '')

    def author(self, default=''):
        """Get post author as a string from post meta information.

        :param default: Default value if author wasn't specified
        :type default: str
        :return: Value of the author field
        :rtype: str

        .. todo::
            Get author/s for pages too.

        .. todo::
            If more than one author, return a list with all authors
            instead of a string, or return always a list and process it
            later.  Use :func:`_parse_list`?
        """
        vs = {'author', 'authors'}
        return self._parse_str_raw(vs, default, ', ')

    def email(self, default=''):
        """Get post author's email as a string from meta information.

        :param default: Default value if author email wasn't specified
        :type default: str
        :return: Value of the email field
        :rtype: str
        """
        vs = {'email', 'e-mail'}
        return self._parse_str_raw(vs, default, ', ')

    def language(self, default='en'):
        """Get post language code as a string from meta information.

        :param default: Default value if language field is empty
        :type default: str
        :return: Value of the language field
        :rtype: str
        """
        vs = {'language', 'lang'}
        return self._parse_str_raw(vs, default, ', ')

    def copyright(self, default=''):
        """Get post copyright as a string from post meta information.

        :param default: Default value if title is empty
        :type default: str
        :return: Value of the copyright field
        :rtype: str
        """
        vs = {'copyright', 'license'}
        return self._parse_str_md(vs, default, ' ')

    def comments(self, default=True):
        """Get comments boolean status.

        :param default: Default value if comments tag is empty
        :type default: str
        :return: ``True`` if entry allows comments, or otherwise
        :rtype: bool
        """
        vs = {'comments'}
        return self._parse_bool(vs, default)

    def private(self, default=False):
        """Get private boolean status.

        :param default: Default value if private tag is empty
        :type default: str
        :return: ``True`` if entry is private, or ``False`` otherwise
        :rtype: bool
        """
        vs = {'private', 'priv'}
        return self._parse_bool(vs, default)

    def navigation(self, default=True):
        """Get navigation boolean status.

        :param default: Default value if navigation tag is empty
        :type default: str
        :return: ``False`` if entry is not in nav. bar, or otherwise
        :rtype: bool
        """
        vs = {'navigation', 'nav'}
        return self._parse_bool(vs, default)

    def tag_list(self, default=[]):
        """Get post tags as a list from post meta information.

        :return: List of tags
        :rtype: list
        """
        vs = {'tags'}
        return self._parse_list(vs, default)

    def tags(self):
        """Retrieve post tags as a string of comma-separated-values.

        :return: List of tags
        :rtype: str
        """
        return self._parse_cvs(self.tag_list())

    def odate_info(self):
        """Get post date as a ``datetime`` object.

        :return: Date field as object
        :rtype: datetime.datetime
        """
        vs = {'odate', 'date', 'created'}
        return self._parse_date_obj(vs)

    def odate(self, date_format='%c'):
        """Get post date as a string.

        :param date_format: Date format string
        :type date_format: str
        :return: Date field as string
        :rtype: str

        .. versionchanged:: 1.3.1a3
            Validate source input meta tags of date raising a
            ``ValueError`` and passing to the logger the information of
            the error.
        """
        try:
            if not self.odate_info() and self.odate_required:
                raise ValueError
        except ValueError:
            self.logger and self.logger.error(
                'Empty or invalid date value in "{}" metadata'.format(
                    self.filename))
            sys.exit(34)
        else:
            return self.odate_info().strftime(date_format) \
                if self.odate_info() else ''

    def mdate_info(self):
        """Get updated post date as a ``datetime`` object.

        :return: Date field as object
        :rtype: datetime.datetime
        """
        vs = {'mdate', 'update', 'updated', 'modified'}
        return self._parse_date_obj(vs)

    def mdate(self, date_format='%c'):
        """Get updated post date as a string.

        :param date_format: Date format string
        :type date_format: str
        :return: Date field as string
        :rtype: str
        :raise ValueError: If data is malformed, but ignored

        .. note:: This is not a mandatory piece of metainformation, so
            it's only taken into consideration when it's well formed and
            doesn't raise a ``ValueError``.
        """
        try:
            if not self.mdate_info():
                raise ValueError
        except ValueError:
            pass
        else:
            return self.mdate_info().strftime(date_format)

    def _is_valid(self):
        """Check if the metadata is valid.

        The metadata is considered malformed when the title (and if
        ``odate_required`` is set to ``True``, also the date) are not in
        the dictionary that is generated after parsing the file.  This
        could happen because:

        * title (or date) are missing;
        * title (or date) identifiers have invalid characters (the only
          valid ones are the ASCII character set encoded: "[A-Za-z0-9]");
        * title (or date) meta-information tag is more than one line
          long and it's not properly indented (subsequent lines require
          to be indented a minimum of four spaces).

        :return: ``False`` if problem found while getting the metadata
        :rtype: bool
        :raise ValueError: If the metadata is malformed

        .. versionadded:: 1.3.1a3
            Preliminary test to see if the mandatory meta tags are set,
            or otherwise an exception is thrown.
        """
        # Title is always required
        title_k = {'title'}
        try:
            if not or_array_in(self.meta, *title_k):
                raise ValueError
        except ValueError:
            self.logger and self.logger.error(
                'Missing or malformed title key in "{}" metadata'.format(
                    self.filename))
            sys.exit(31)

        # Date is required only in posts, but not in pages
        if self.odate_required:
            odate_k = {'odate', 'date', 'created'}
            try:
                if not or_array_in(self.meta, *odate_k):
                    raise ValueError
            except ValueError:
                self.logger and self.logger.error(
                    'Missing or malformed date key in "{}" metadata'.format(
                        self.filename))
                sys.exit(32)

    def _parse_str_raw(self, values, default='', joint=', '):
        """Get a value as a raw string from post meta information.

        Raw, in this context, means in plain text, with no markup
        language.

        :param values: Set of possible meta tags
        :type values: set
        :param default: Default value if no keys match
        :type default: str
        :param joint: String used as a joint between values
        :type joint: str
        :return: The value of the meta key, or the default value
        :rtype: str
        """
        value = or_array_in(self.meta, *values)
        if value:
            return joint.join(value)
        else:
            return default

    def _parse_str_md(self, values, default='', joint=' ', period=False):
        """Get value as string from metadata allowing Markdown syntax.

        It parses the Markdown code to HTML and allow this HTML for this
        particular piece of metadata, except some block elements such as
        ``p``, ``br``...

        :param values: Set of possible meta tags
        :type values: set
        :param default: Default value if no keys match
        :type default: str
        :param joint: String used as a joint between values
        :type joint: str
        :param period: Add a period at the end, if there's none
        :type period: bool
        :return: The value of the meta key, or the default value
        :rtype: str
        """
        value = or_array_in(self.meta, *values)
        if value:
            fmt = markdown.markdown(joint.join(value))
            parsed_str = re.sub(r'</*(p|br)[^>]*?>', '', fmt)
        else:
            parsed_str = default

        if period and parsed_str and not parsed_str.endswith('.'):
            parsed_str = parsed_str + '.'

        return parsed_str

    def _parse_bool(self, values, default=True):
        """Get a value as a boolean from post meta information.

        The function has two "confrontation arrays (sets)" to test
        against them the values of the intended key.  Because the meta
        tag possible values are "yes" or "no" in natural language, this
        lists have all the possible outcome, and uses one list or the
        other depending on the default value.

        :param values: Set of possible meta tags
        :type values: set
        :param default: Default value if no keys match
        :type default: str
        :return: The value of the meta key, or the default value
        :rtype: str
        """
        if default:
            bools = {'no', 'non', 'não', 'nao', 'ne', 'nein',
                     'false', '0'}
        else:
            bools = {'yes', 'sí', 'si', 'sì', 'sim', 'jes', 'oui', 'ja'
                     'true', '1'}

        value = or_array_in(self.meta, *values)
        if value:
            if any(x in map(lambda x: x.lower(), value) for x in bools):
                return not default
        return default

    def _parse_list(self, values, default=[]):
        """Get a list with all the values to the matching key.

        :param values: Set of possible meta tags
        :type values: set
        :param default: Default value if no keys match
        :type default: str
        :return: The value of the meta key as a list, or default value
        :rtype: list
        """
        value = or_array_in(self.meta, *values)
        if value:
            return [x.strip() for x in str(value[0]).split(',') if x.strip()]
        else:
            return default

    def _parse_csv(self, values, default=''):
        """Get string with values to the matching key, comma separated.

        :param values: Set of possible meta tags
        :type values: set
        :param default: Default value if no keys match
        :type default: str
        :return: The value of the meta key as a string, or default value
        :rtype: string
        """
        return ', '.join(self._dt_parse_list(values, default))

    def _parse_date_obj(self, values):
        """Get a ``datetime`` object from the metadata.

        :param values: Set of possible meta tags
        :type values: set
        :return: Datetime object
        :rtype: datetime.datetime
        """
        value = or_array_in(self.meta, *values)
        return dt_parse(''.join(value)) if value else None

    def __repr__(self):
        """Return dictionary with all metainformation."""
        return self.as_dict()

    def __str__(self):
        """Return string representation for an object of this class."""
        return str(self.meta)



#!/usr/bin/env python3
# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Main entry point.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT
"""
from pynfact.cli import arg_build
from pynfact.cli import arg_init
from pynfact.cli import arg_serve
from pynfact.cli import set_logger
import argparse
import sys


try:
    import colored_traceback.auto
    colored_traceback.add_hook(always=True)
except ImportError:
    pass


def main():
    """Manage the command line arguments.

    .. versionchanged:: 1.2.0a1
        Implement ``logging`` instead of printing to ``stdout`` and/or
        ``stderr``.

    .. versionchanged:: 1.3.1a1
        Retrieve arguments by using :mod:`argparse`.

    .. versionchanged:: 1.3.1a2
        Move functions to :mod:`cli` and renamed file as ``main.py``.
    """
    parser = argparse.ArgumentParser(description=""
                                     "PynFact!: "
                                     "Static website generator from Markdown "
                                     "and reStructuredText to HTML5",
                                     prog="pynfact",)
    # formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    rgroup = parser.add_mutually_exclusive_group(required=True)

    rgroup.add_argument('-i', '--init', default=None,
                        metavar='<site>',
                        help="initialize a new website structure")
    rgroup.add_argument('-b', '--build', action='store_true',
                        help="parse input files and build the website")
    rgroup.add_argument('-s', '--serve', nargs='?',
                        default='localhost', const='localhost',
                        metavar='<host>',
                        help="set host where to serve the website (localhost)")
    parser.add_argument('-p', '--port', default=4000,
                        metavar='<port>', type=int,
                        help="set port to listen to when serving (4000)")
    parser.add_argument('-l', '--log', default='pynfact.err',
                        metavar='<logfile>',
                        help="set file where to log errors (pynfact.err)")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="increase output verbosity")

    # Print help if no arguments supplied
    args = parser.parse_args(None if sys.argv[1:] else ['--help'])

    # Process arguments
    logger = set_logger(args.verbose, error_log=args.log)
    if args.init:
        arg_init(logger, args.init)
    elif args.build:
        arg_build(logger)
    elif args.serve:
        arg_serve(logger, args.serve, int(args.port))


# Main entry
if __name__ == '__main__':
    main()



# vim: set ft=python fileencoding=utf-8 tw=72 fdm=indent nowrap:
"""
Build the content.

:copyright: © 2012-2020, J. A. Corbal
:license: MIT

.. versionchanged:: 1.3.1a4
    Include processing of reStructuredText files along with Markdown.
    Those files are detected by extension and parsed accordingly.  So,
    no extension test in this module, only in the parser.

.. versionchanged:: 1.3.1b4
    Replaced functions ``_make_entry_uri`` and ``_make_root_uri`` with
    just :func:``_make_uri``, which will behave like the former if the
    argument ``for_entry`` is set to ``True``, and the latter when set
    to ``False``.
"""
from datetime import datetime
from feedgen.feed import FeedGenerator
from jinja2 import Environment, FileSystemLoader, Markup
from math import ceil
from pynfact.fileman import link_to
from pynfact.fileman import has_extension_md_rst
from pynfact.meta import Meta
from pynfact.parser import Parser
from pynfact.struri import slugify, strip_html_tags
import distutils.dir_util
import filecmp
import gettext
import locale
import os
import resource
import sys


class Builder:
    """Site building process manager.

    .. todo::
        Manage better the timezone, the default is always UTC, for posts
        info, and both feeds publication and modification dates.

    .. versionchanged:: 1.0.1.dev1
        Constructor takes a configuration dictionary instead of all
        parameters individually.

    .. versionchanged:: 1.0.1a1
        Input pages stored in ``pages`` directory instead of root.

    .. versionchanged:: 1.2.0a1
        Implement ``logging`` instead of printing to ``stdout`` and/or
        ``stderr``.

    .. versionchanged:: 1.3.1a4
        Allowed extension in constructor is no longer needed.  Input
        files are processed depending on their extension (Markdown or
        reStructuredText), and those are fixed.

    .. versionchanged:: 1.3.1b1
        Test of extension done by :func:`has_extension_md_rst` from
        module :mod:`fileman`.  No extensions passed as parameters nor
        defined explicitly in this class.

    .. versionchanged:: 1.3.1b2
        Feed can be disabled and will not be generated if the value of
        ``feed_format`` is neither "rss" nor "atom" (case insensitive).

    .. versionchanged:: 1.3.1b3
        Pages and entries files are read only once, put in memory and
        work from there without touching the files anymore.
    """

    def __init__(self, site_config, template_values=dict(), logger=None):
        """Constructor.

        :param config: Site configuration as multidimensional dictionary
        :type config: dict
        :param template_values: Common values in all templates
        :type template_values: dict
        :param logger: Logger where to store activity in
        :type logger: logging.Logger
        :raise localeError: If the selected locale is not supported
        """
        self.site_config = site_config
        self.template_values = template_values
        self.site_config['dirs']['deploy'] = \
            os.path.join(self.site_config.get('dirs').get('deploy'),
                         self.site_config.get('uri').get('base'))
        self.logger = logger

        # Set locale for the site.
        self.old_locale = locale.getlocale()
        try:
            self.current_locale = \
                locale.setlocale(locale.LC_ALL,
                                 self.site_config.get('wlocale').get('locale'))
        except locale.Error:
            self.logger and self.logger.error(
                "Unsupported locale setting")
            sys.exit(41)

        # Constants-like (I don't like this approach)
        # source dirs.
        self_dir = os.path.dirname(os.path.realpath(__file__))
        self.locale_dir = os.path.join(self_dir, 'data/locale')
        self.templates_dir = 'templates'
        self.builtin_templates_dir = \
            os.path.join(self.site_config.get('dirs').get('deploy'),
                         self.templates_dir)

        # dest. dirs.
        self.home_cont_dir = 'page'         # Home paginator
        self.archive_dir = 'archive'        # Archive list page
        self.categories_dir = 'categories'  # One dir. per category
        self.tags_dir = 'tags'              # One dir. per tag

        # src. & dest. dirs.
        self.pages_dir = 'pages'            # Other pages such as 'About'...
        self.entries_dir = 'posts'          # Where posts are
        self.static_dir = 'static'          # CSS, JS, &c.

        # Default values to override meta information
        self.meta_defaults = {
            'author':
                self.site_config.get('info').get('site_author'),
            'category':
                self.site_config.get('presentation').get('default_category'),
            'language':
                self.site_config.get('wlocale').get('language'),
        }

        # Generate all entries metadata, once
        self.entries_dict = self._gather_content_data().get('entries')
        self.pages_dict = self._gather_content_data().get('pages')

    def __del__(self):
        """Destructor.  Restore the locale, if changed.

        At the end of the site generation, the destructor prints the
        maximum resident set size used of this class, if the logger is
        set to show debug messages.   That is, the maximum number of
        kilobytes of physical memory that processes used simultaneously.
        """
        if sys.platform.startswith('linux') or \
                sys.platform.startswith('freebsd') or \
                sys.platform.startswith('openbsd'):
                # On Linux the result is in kilobytes.
            units = 'kilobytes'
            peak = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        elif sys.platform == 'darwin':
            # On OSX the result is in bytes.
            units = 'bytes'
            peak = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        else:
            units = ''
            peak = '(unsupported)'

        self.logger and self.logger.debug(
            'Maximum resident set size used: {} {}'.format(peak, units))

        return locale.setlocale(locale.LC_ALL, self.old_locale)

    def gen_entry(self, filename, date_format='%c'):
        """Generate a HTML entry from its markup language counterpart.

        :param filename: Markdown or reStructuredText file to parse
        :type filename: str
        :param date_format: Date format for entry
        :type date_format: str
        :return: Generated HTML
        :rtype: str
        """
        override = {'content': self._fetch_html(self.entries_dir, filename)}
        self.entries_dict.get(filename).update(override)
        values = self.template_values.copy()
        values['entry'] = self.entries_dict.get(filename)
        self._update_meta_date_format(values.get('entry'), date_format)
        outfile = self._make_output_file(
            values.get('entry').get('title'),
            self._entry_link_prefix(filename))

        return self._render_template('entry.html.j2', outfile, values)

    def gen_entries(self, date_format='%c'):
        """Generate all entries.

        :param date_format: Date format for entry
        :type date_format: str
        """
        for filename, meta in self.entries_dict.items():
            self.gen_entry(filename, date_format=date_format)

    def gen_home(self, max_entries_per_page=10, date_format='%Y-%m-%d'):
        """Generate home page, and subpages.

        :param max_entries: Max. entries per page
        :type max_entries: int
        :param date_format: Date format for home page
        :type date_format: str
        """
        entries = list()
        total_entries = 0

        for filename, meta in self.entries_dict.items():
            override = {'uri': self._make_uri(meta.get('title'),
                                              filename, for_entry=True,
                                              absolute=True)}
            self.entries_dict[filename].update(override)
            self._update_meta_date_format(meta, date_format)
            if not meta['private']:
                total_entries += 1
                entries.append(self.entries_dict[filename])

        entries = sorted(entries, key=lambda k: k.get('odate_idx'),
                         reverse=True)

        # Paginator
        total_pages = ceil(total_entries / max_entries_per_page)
        values = self.template_values.copy()

        # Generate 'index.html' even when there are no posts
        if total_entries == 0:
            outfile = self._make_output_file()
            self._render_template('entries.html.j2', outfile, values)

        # Home page (and subsequent ones)
        for cur_page in range(1, total_pages + 1):
            min_page = (cur_page - 1) * max_entries_per_page
            max_page = cur_page * max_entries_per_page

            values['entries'] = entries[min_page:max_page]
            values['cur_page'], values['total_pages'] = \
                cur_page, total_pages

            if cur_page == 1:
                # This is the home page, the "index.html" of the site
                outfile = self._make_output_file()
            else:
                # These are the subsequent pages other than the first
                outfile = self._make_output_file(
                    str(cur_page), self.home_cont_dir)

            self._render_template('entries.html.j2', outfile, values)
            values['entries'] = {}  # reset the entries dict.

    def gen_archive(self, date_format='%c'):
        """Generate complete website archive, based on date.

        :param date_format: Date format for entry
        :type date_format: str
        """
        archive = dict()
        for filename, meta in self.entries_dict.items():
            override = {'uri': self._make_uri(meta.get('title'),
                                              filename, for_entry=True,
                                              absolute=False)}
            self.entries_dict.get(filename).update(override)
            if not meta.get('private'):
                if meta.get('oyear_idx') in archive:
                    archive[meta.get('oyear_idx')].setdefault(
                        meta.get('omonth_idx'), []).append(meta)

                    # sort entries
                    archive[meta.get('oyear_idx')][meta.get('omonth_idx')] = \
                        sorted(archive[meta.get('oyear_idx')][meta.get('omonth_idx')],
                               key=lambda k: k.get('odate_idx'),
                               reverse=False)
                else:
                    archive[meta.get('oyear_idx')] = dict()
                    archive[meta.get('oyear_idx')][meta.get('omonth_idx')] = \
                        [meta]

        values = self.template_values.copy()
        values['archive'] = archive
        outfile = self._make_output_file('', self.archive_dir)
        return self._render_template('archive.html.j2', outfile, values)

    def gen_category_list(self, date_format='%c'):
        """Generate categories page (an archive sorted by category).

        :param date_format: Date format for entry
        :type date_format: str
        """
        categories = dict()
        for filename, meta in self.entries_dict.items():
            self._update_meta_date_format(meta, date_format)
            override = {'uri': self._make_uri(meta.get('title'),
                                              filename, for_entry=True,
                                              absolute=False)}
            self.entries_dict.get(filename).update(override)
            if not meta.get('private'):
                categories.setdefault(
                    meta.get('category'), []).append(meta)

                # sort entries
                categories[meta.get('category')] = \
                    sorted(categories[meta.get('category')],
                           key=lambda k: k.get('odate_idx'),
                           reverse=True)

        values = self.template_values.copy()
        values['categories'] = categories
        outfile = self._make_output_file('', self.categories_dir)
        return self._render_template('catlist.html.j2', outfile, values)

    def gen_categories(self, date_format='%c'):
        """Generate categories pages.

        :param date_format: Date format for entry
        :type date_format: str
        """
        entries_by_category = dict()
        for filename, meta in self.entries_dict.items():
            self._update_meta_date_format(meta, date_format)
            if not meta.get('private'):
                self._update_meta_date_format(meta, date_format)
                entries_by_category.setdefault(
                    meta.get('category'), []).append(meta)

        # One page for each category
        values = self.template_values.copy()
        for category in entries_by_category:
            values['category_name'] = category
            values['entries'] = entries_by_category.get(category)

            # sort entries
            values['entries'] = sorted(values.get('entries'),
                                       key=lambda k: k.get('odate_idx'),
                                       reverse=True)
            outfile = self._make_output_file(category, self.categories_dir)
            self._render_template('cat.html.j2', outfile, values)

    def gen_tags(self, date_format='%c'):
        """Generate tags pages.

        :param date_format: Date format for entry
        :type date_format: str
        """
        entries_by_tag = dict()
        for filename, meta in self.entries_dict.items():
            if not meta.get('private'):
                for tag in meta.get('tag_list'):
                    self._update_meta_date_format(meta, date_format)
                    entries_by_tag.setdefault(tag, []).append(meta)

        # One page for each tag
        values = self.template_values.copy()
        for tag in entries_by_tag:
            values['tag_name'] = tag
            values['entries'] = entries_by_tag.get(tag)
            # sort entries
            values['entries'] = sorted(values.get('entries'),
                                       key=lambda k: k.get('odate_idx'),
                                       reverse=True)
            outfile = self._make_output_file(tag, self.tags_dir)
            self._render_template('tag.html.j2', outfile, values)

    def gen_tag_cloud(self):
        """Generate tags cloud page.

        Tags will appear in different sizes depending on the their
        occurrences along the posts.  The more a tag is used, the bigger
        will be displayed.
        """
        entries_by_tag = dict()
        for filename, meta in self.entries_dict.items():
            override = {'uri': self._make_uri(meta.get('title'),
                                              filename, for_entry=True,
                                              absolute=False)}
            self.entries_dict.get(filename).update(override)
            if not meta.get('private'):
                for tag in meta.get('tag_list'):
                    entries_by_tag.setdefault(tag, []).append(meta)

        # Multipliers seq. for tag size in function of times repeated
        tagcloud_seq = [0, 14, 21, 27, 32, 38, 42, 45, 47, 48, 50, 52]

        # One page for each tag
        values = self.template_values.copy()
        values['tags'] = list()
        for tag in entries_by_tag:
            if tag:
                tagfreq = len(entries_by_tag.get(tag))
                mult = 100 + int(tagcloud_seq[-1]
                                 if tagfreq > len(tagcloud_seq)
                                 else tagcloud_seq[tagfreq - 1])
                values.get('tags').append({tag: mult})

        outfile = self._make_output_file('', self.tags_dir)
        self._render_template('tagcloud.html.j2', outfile, values)

    def gen_nav_page_links(self):
        """Update the template data to contain also all page links.

        The navigation bar contains, not all links relevant to the site,
        but also all pages without the meta descriptor ``Navigate`` set
        to "no".

        .. important::
            Since this add new data to the base template, it **must be
            invoked first**, before generating any other content.
        """
        for filename, meta in self.pages_dict.items():
            if meta.get('navigation') and not meta.get('private'):
                values = self.template_values.copy()
                values['page'] = self.pages_dict.get(filename)

                # Update base template values to get pages links in
                # nav. bar, only for those pages that have the meta tag
                # ``navigation`` set to true
                self.template_values['blog']['page_links'].append(
                    [values.get('page').get('title'),
                     values.get('page').get('uri')])

                self.logger and self.logger.info(
                    'Added page to navigation: "{}"'.format(filename))

    def gen_page(self, filename):
        """Generate a HTML page from its markup language counterpart.

        :param filename: Markdown or reStructuredText file to parse
        :type filename: str
        :return: Generated HTML
        :rtype: str
        """
        override = {'content': self._fetch_html(self.pages_dir, filename)}
        self.pages_dict.get(filename).update(override)
        values = self.template_values.copy()
        values['page'] = self.pages_dict.get(filename)
        outfile = self._make_output_file(values.get('page').get('title'))
        return self._render_template('page.html.j2', outfile, values)

    def gen_pages(self):
        """Generate all pages."""
        for filename, meta in self.pages_dict.items():
            self.gen_page(filename)

    def gen_feed(self, feed_format="atom", outfile='feed.xml'):
        """Generate blog feed.

        :param feed_format: Feed format string ('rss' or 'atom').
                            If invalid value, will default to 'atom'
        :type feed_format: str
        :param outfile: Output filename
        :type outfile: str
        """
        if feed_format.lower() != "atom" and \
           feed_format.lower() != "rss":
            return

        feed = FeedGenerator()
        # feed.logo()
        feed.id(self.site_config.get('info').get('site_name')
                if self.site_config.get('info').get('site_name')
                else self.site_config.get('uri').get('canonical'))
        feed.title(self.site_config.get('info').get('site_name')
                   if self.site_config.get('info').get('site_name')
                   else self.site_config.get('uri').get('canonical'))
        feed.subtitle(self.site_config.get('info').get('site_description')
                      if self.site_config.get('info').get('site_description')
                      else 'Feed')
        feed.author({
            'name':
                self.site_config.get('info').get('site_author'),
            'email':
                self.site_config.get('info').get('site_author_email')
        })
        feed.description(self.site_config.get('info').get('site_description'))
        feed.link(href=os.path.join(
            self.site_config.get('uri').get('canonical'),
            self.site_config.get('uri').get('base')),
            rel='alternate')
        feed.link(href=os.path.join(
            self.site_config.get('uri').get('canonical'),
            self.site_config.get('uri').get('base'),
            outfile), rel='self')
        feed.language(self.site_config.get('wlocale').get('language'))
        feed.copyright(self.site_config.get('info').get('copyright'))

        entries = list()
        for filename, meta in self.entries_dict.items():
            if not meta.get('private'):
                uri = self._make_uri(meta.get('title'), filename,
                                     for_entry=True)
                override = {
                    'content':
                        self._fetch_html(self.entries_dir, filename),
                    'full_uri':
                        os.path.join(
                            self.site_config.get('uri').get('canonical'),
                            self.site_config.get('uri').get('base'),
                            uri),
                }
                self.entries_dict.get(filename).update(override)
                entries.append(meta)

        # Sort chronologically descent
        entries = sorted(entries, key=lambda k: k.get('odate_idx'),
                         reverse=True)
        for entry in entries:
            fnew = feed.add_entry()
            fnew.id(slugify(strip_html_tags(entry.get('title'))))
            fnew.title(entry.get('title'))
            fnew.description(entry.get('content'))
            if entry.get('mdate_html'):
                fnew.updated(entry.get('mdate_html'))
                fnew.published(entry.get('mdate_html'))
            else:
                fnew.updated(entry.get('odate_html'))
            fnew.pubDate(entry.get('odate_html'))
            # , 'email':entry.get('email')})
            fnew.author({'name': entry.get('author')})
            fnew.link(href=entry.get('full_uri'), rel='alternate')

        if feed_format.lower() == "rss":
            feed.rss_file(
                os.path.join(
                    self.site_config.get('dirs').get('deploy'), outfile))
        elif feed_format.lower() == "atom":
            feed.atom_file(
                os.path.join(
                    self.site_config.get('dirs').get('deploy'), outfile))

    def gen_static(self):
        """Generate (copies) static directory."""
        src = self.static_dir
        dst = os.path.join(
            self.site_config.get('dirs').get('deploy'),
            self.static_dir)
        if os.path.exists(src):
            distutils.dir_util.copy_tree(src, dst, update=True,
                                         verbose=True)

    def gen_extra_dirs(self):
        """Generate extra directories if they exist."""
        if self.site_config.get('dirs').get('extra'):
            for extra_dir in self.site_config.get('dirs').get('extra'):
                src = extra_dir
                dst = \
                    os.path.join(self.site_config.get('dirs').get('deploy'),
                                 extra_dir)
                if os.path.exists(src):
                    distutils.dir_util.copy_tree(src, dst, update=True,
                                                 verbose=True)

    def gen_site(self):
        """Generate all website content.

        .. note::
            The first thing that has to be generated is the navigation
            links for all user defined pages.  Otherwise those links
            could be left behind on page pages.
        """
        self.logger and self.logger.info('Building static website...')

        self.gen_nav_page_links()
        self.gen_entries(self.site_config.get('date_format').get('entry'))
        self.gen_pages()
        self.gen_archive(self.site_config.get('date_format').get('list'))
        self.gen_categories(self.site_config.get('date_format').get('list'))
        self.gen_category_list(self.site_config.get('date_format').get('list'))
        self.gen_tags(self.site_config.get('date_format').get('list'))
        self.gen_tag_cloud()
        self.gen_home(self.site_config.get('presentation').get('max_entries'),
                      self.site_config.get('date_format').get('home'))
        self.gen_feed(self.site_config.get('presentation').get('feed_format'))
        self.gen_static()
        self.gen_extra_dirs()

    def _gather_content_data(self):
        """Gather all metadata from all parseable files.

        A parseable files is one with Markdown or reStructuredText valid
        extension, as described in :func:`fileman.has_extension_md_rst`.
        This way, the files will be transversed only once, and stored.

        This stores only entries information, not pages metadata, since
        pages only need to be renderer "as is".  Entries, on the other
        side, are also required to extract information in order to make
        category pages, tags, tag clouds, etc.  So, it's needed to have
        all that data in the same place.

        The structure of the return value is as follows::

            data = {
                ['entries']: {
                    'entry1.filename': Meta(entry1).as_dict(),
                    'entry2.filename': Meta(entry2).as_dict(),
                }
                ['pages']: {
                    'page1.filename': Meta(page1).as_dict(),
                    'page2.filename': Meta(page2).as_dict(),
                }
            }

        :return: Dictionary of all parseabe objects and their metadata
        :rtype: dict
        """
        # Gather entries
        entries_dict = dict()
        for filename in os.listdir(self.entries_dir):
            if has_extension_md_rst(filename):
                meta = self._fetch_meta(self.entries_dir, filename,
                                        odate_required=True)

                override_entry = {
                    'category_uri':
                        self._make_uri(meta.category(),
                                       self.categories_dir,
                                       for_entry=False),
                    'odate_idx': meta.odate('%Y-%m-%d %H:%M:%S'),
                    'oyear_idx': meta.odate('%Y'),
                    'omonth_idx': meta.odate('%m-%d'),
                    'site_comments':
                        self.site_config.get('presentation').get('comments'),
                    'uri':
                        self._make_uri(meta.title(),
                                       filename, for_entry=True,
                                       absolute=True),
                }
                entries_dict[filename] = \
                    meta.as_dict(override_entry, self.meta_defaults)

        # Gather pages
        pages_dict = dict()
        for filename in os.listdir(self.pages_dir):
            if has_extension_md_rst(filename):
                meta = self._fetch_meta(self.pages_dir, filename,
                                        odate_required=False)

                override_page = {'uri': self._make_uri(meta.title(),
                                                       for_entry=False)}
                pages_dict[filename] = \
                    meta.as_dict(override_page, self.meta_defaults)

        return {'entries': entries_dict, 'pages': pages_dict}

    def _update_meta_date_format(self, meta, date_format):
        """Update the date format from a meta dictionary object.

        Operates on a dictionary of metainformation, output of
        :fun:`Meta.as_dict``, so the key names are known.  Uses the
        two datetime information fields, ``odate_info`` for the original
        date, and ``mdate_info`` for the modified date, and rewrites in
        place the values of their string counterparts, ``odate`` and
        ``mdate``, if they are not ``None``.

        :param meta: Metainformation dictionary
        :type meta: dict
        :param date_format: New date format for ``odate`` and ``mdate``
        :type date_format: str

        .. seealso:: :fun:`Meta.as_dict`
        """
        if meta.get('odate'):
            meta['odate'] = meta.get('odate_info').strftime(date_format)

        if meta.get('mdate'):
            meta['mdate'] = meta.get('mdate_info').strftime(date_format)

    def _render_template(self, template, output_data, values):
        """Render a template using Jinja2.

        :param template: Template to use
        :type template: str
        :param output_data: File where the data is saved
        :type output_data: str
        :return: Generated HTML of the output data
        :rtype: str
        """
        trans = gettext.translation('default', self.locale_dir,
                                    [self.current_locale])
        env = Environment(extensions=['jinja2.ext.i18n'],
                          loader=FileSystemLoader(
                              [self.templates_dir,
                               self.builtin_templates_dir]))
        env.install_gettext_translations(trans)
        env.globals['slugify'] = slugify  # Add `slugify` to Jinja2
        env.globals['strip_html_tags'] = strip_html_tags
        template = env.get_template(template)
        html = template.render(**values)

        # Update only those files that are different in content
        # comparing with a cache file
        with open(output_data + '~', mode="w",
                  encoding=self.site_config.get('wlocale').get('encoding')) \
                as cache_file:
            cache_file.write(html)

        if not os.path.exists(output_data) or \
           not filecmp.cmp(output_data + '~', output_data):
            with open(
                output_data, mode="w",
                encoding=self.site_config.get('wlocale').get('encoding')) \
                    as output_file:
                output_file.write(html)
            self.logger and self.logger.info(
                'Updated content of: "{}"'.format(output_data))

        # Clear cache, both in memory and space
        filecmp.clear_cache()
        os.remove(output_data + '~')

        return html

    def _fetch_markup(self, directory, filename):
        """Parse an input file depending on its extension.

        The ``Parser`` constructor detects the extension of ``filename``
        and parses its body content as well as its metadata.

        :param directory: Path to the file to be parsed
        :type directory: str
        :param filename: File to parse
        :type filename: str
        :return: Parser object
        :rtype: Parser

        .. sealso:: :class:`Parser`
        """
        return Parser(os.path.join(directory, filename),
                      encoding=self.site_config.get('wlocale').get('encoding'),
                      logger=self.logger)

    def _fetch_html(self, directory, filename):
        """Fetch HTML content out of a markup language input file.

        :param directory: Markdown or reStructuredText file directory
        :type directory: str
        :param filename: Markdown or reStructuredText file to parse
        :type filename: str
        :return: HTML content for the parsed file
        :rtype: str
        """
        return self._fetch_markup(directory, filename).html()

    def _fetch_meta(self, directory, filename, odate_required=False):
        """Fetch metadata out of a markup language input file.

        If ``odate_required`` is ``False``, the metadata is valid as
        long as it has a title; if it's ``True``, the metadata require
        to be valid, also an "origin date", or ``odate`` field.

        :param directory: Markdown or reStructuredText file directory
        :type directory: str
        :param filename: Markdown or reStructuredText file to parse
        :type filename: str
        :param odate_required: Field ``odate`` is required
        :type odate_required: bool
        :return: Parsed file metadata
        :rtype: Meta
        """
        return Meta(self._fetch_markup(directory, filename).metadata(),
                    filename, odate_required, logger=self.logger)

    def _entry_link_prefix(self, entry):
        """Compute entry final path.

        To compute the final path, it's required to take the meta
        information of that entry concerning to title and origin date.
        Let's suppose the title is "My post", and the ``odate`` set to
        "2020-04-01".  The output will be: ``posts/2020/04/01/my-post``.

        :param: Entry filename
        :type: str
        :return: Path to this entry relative to root (slugified)
        :rtype: str

        .. versionchanged:: 1.3.0a
            Set as a member method.
        """
        meta = self._fetch_meta(self.entries_dir, entry,
                                odate_required=True)

        category = meta.category(
            self.site_config.get('presentation').get('default_category'))
        odate = meta.odate('%Y-%m-%d')
        odate_arr = odate.split('-')
        path = os.path.join(str(self.entries_dir),
                            slugify(category),
                            str(odate_arr[0]),
                            str(odate_arr[1]),
                            str(odate_arr[2]))
        return path

    def _make_uri(self, name='', infix='', index='index.html',
                  for_entry=True, absolute=False):
        """Generate the link to an entry, based on the date and name.

        Link automated generation for final URIs of articles.

        :type name: Destination basename
        :param name: str
        :param infix: Appendix to the deploy directory
        :type infix: str
        :param index: Default name of the generated resource
        :type index: str
        :param absolute: Force the URI to be done from the root
        :type absolute: bool
        :param for_entry: Entry URI if ``True``, else make it from root
        :type for_entry: bool
        :return: Link to final external URI relative to root directory
        :rtype: str

        .. versionchanged:: 1.3.1b4
            Make a link either to a blog entry or to a page in just one
            method, instead of using ``_make_entry_uri`` and
            ``_make_root_uri``.
        """
        if for_entry:
            if absolute:
                path = os.path.join('/',
                                    self.site_config.get('uri').get('base'),
                                    self._entry_link_prefix(infix))

            else:
                path = self._entry_link_prefix(infix)
        else:
            path = os.path.join('/',
                                self.site_config.get('uri').get('base'),
                                infix)

        return link_to(name, path, makedirs=False, justdir=True,
                       index=index)

    def _make_output_file(self, name='', infix='', index='index.html'):
        """Return the output file link, and make required directories.

        Link automated generation for files in the deploy directory.

        :param name: Destination basename
        :type name: str
        :param infix: Appendix to the deploy directory
        :type infix: str
        :param index: Default name of the generated resource
        :type index: str
        :return: Link to the constructed path in the deploy directory
        :rtype: str
        """
        return link_to(name,
                       os.path.join(self.site_config.get('dirs').get('deploy'),
                                    infix),
                       makedirs=True, justdir=False, index=index)
